/*
 * (C) Copyright Architecture and Language Implementation Laboratory,
 * Department of Computer Science,
 * University of Massachusetts at Amherst. 2001
 * All rights reserved.
 */

package com.ibm.JikesRVM;

import com.ibm.JikesRVM.classloader.*;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.io.FileReader;
import java.io.LineNumberReader;
import java.util.List;
import java.util.ArrayList;

/**
 * Utility to read compiler advice annotations from file in ascii format.
 * Takes a single argument: the name of the file containing the ascii
 * annotations.  Each line of the file corresponds to an annotation
 * for one method and has the following format:
 * <p>
 * <pre>
 * <class> <method> <signature> <advice> <optLevel>
 * </pre>
 * Where the types and meanings of the fields is as follows:
 * <ul>
 * <li><code>&lt;class></code> <i>string</i> The name of the class</li>
 * <li><code>&lt;method></code> <i>string</i> The name of the method</li>
 * <li><code>&lt;signature></code> <i>string</i> The method signature</li>
 * <li><code>&lt;advice></code> <i>int</i> The compiler type to be used --
 * an integer value corresponding to the compiler enumeration in 
   VM_CompiledMethod</li>
 * <li><code>&lt;optLevel></code> <i>int</i> (Optional) The opt level to use 
   if compiler is optimizing compiler</li>
 * </ul>
 *
 * @author: Chris Hoffmann
 * @version $Revision: 1.1 $
 * @date $Date: 2004/09/07 13:42:44 $
 *
 * @see VM_CompilerAdvice
 * @see VM_CompilerAdviceAttribute
 * @see VM_CompilerInfo
 */
class VM_CompilerAdviceInfoReader {

  /**
   * Read annoations from a specified file. Reads all annotations at
   * once and returns a collection of compiler advice attributes.
   * 
   * @param file The annoation file to be read
   * @return A list of compileration advice attributes
   */
  public static List readCompilerAdviceFile(String file) {
    List compilerAdviceInfo = null;
    LineNumberReader fileIn = null;
    
    if (file == null) return null;
    
    try {
      fileIn = new LineNumberReader(new FileReader(file));
    } catch (java.io.FileNotFoundException e) {
      System.out.println("IO: Couldn't read compiler advice attribute file: " 
			 + file + e);
      return null;
    }
    
    StreamTokenizer st = new StreamTokenizer(fileIn);
    
    // Before reading attributes change the syntax of tokenizer for
    // correctly reading method descriptors
    
    // recognize single line Java comments
    st.slashSlashComments(true); // recognize java double slash comments
    st.eolIsSignificant(true);   // end of lines are significant
    st.wordChars('(', ')');   // parens are part of method desc.
    st.wordChars('/', '/');   // slashes are part of method desc.
    st.wordChars('[', '[');   // so are array chars [
    st.wordChars(';', ';');   // and semicolons
    
    st.wordChars('<', '<');   // angle brackets are part of
    st.wordChars('>', '>');   // <init> and <clinit> special methods
    
    st.wordChars('_', '_');   // underscore may occur in identifiers
    st.wordChars('$', '$');   // dollar symbol may occur for inner classes
    
    try {
      compilerAdviceInfo = readCompilerAdvice(st);
      fileIn.close();
    } catch (IOException e) {
      System.out.println("IO: Couldn't read attribute file: " + file + e );
      
    } catch (Exception badFormat) {
      System.out.println("Incorrect attribute format for  file: " + file
			 + " line: " + fileIn.getLineNumber() + " " 
			 +  badFormat);
      badFormat.printStackTrace();
    }
    
    return compilerAdviceInfo;
  }
  
  /**
   * Reads all annotations at once from a stream and returns a
   * collection of compiler advice attributes.
   *
   * @param in An input stream
   * @return A collection (<code>List</code>) of per-site compiler advice
   */
  private static List readCompilerAdvice(StreamTokenizer in)
	throws IOException, Exception {
    List compilerAdviceInfo = new ArrayList();
    
    
    // iterate over the attributes
    while (in.nextToken() != StreamTokenizer.TT_EOF) {
      if (in.ttype == StreamTokenizer.TT_EOL) continue;
      
      // return the token read  for the while loop, it will be read again by
      // readOneAttribute
      in.pushBack();
      compilerAdviceInfo.add(readOneAttribute(in));
    }
    
    return compilerAdviceInfo;
  }

  /**
   * Actual reading is done here.  This method reads one attribute
   * from a single line of an input stream.  There are six elements
   * per line corresponding to each call site. First three are
   * strings, <i>class name</i>, <i>method name</i>, <i>method
   * signature</i>, followed by one number, 
   * <i>compiler advice</i>.
   *
   * @param st an input stream
   * @return an compileration advice atribute
   */
  private static VM_CompilerAdviceAttribute readOneAttribute(StreamTokenizer st)
    throws IOException, Exception {
    String className, methodName, methodSignature;
    int compiler, optLevel = -1;
    
    // first token of an attribute must be class name
    int tokenType = st.nextToken();
    if (tokenType != StreamTokenizer.TT_WORD)
      throw new Exception("Expecting class name: " + st.ttype);
    className = st.sval;
    
    // read method name
    tokenType = st.nextToken();
    if (tokenType != StreamTokenizer.TT_WORD)
      throw new Exception("Expecting method name: " + st.ttype);
    methodName = st.sval;
    
    // read method signature
    tokenType = st.nextToken();
    if (tokenType != StreamTokenizer.TT_WORD)
      throw new Exception("Expecting method signature: " + st.ttype);
    methodSignature = st.sval;
    
    // read compiler type
    tokenType = st.nextToken();
    if (tokenType != StreamTokenizer.TT_NUMBER)
      throw new Exception("must be compiler type (int)" + st);
    compiler = (int)st.nval;
    
     // read optLevel type
    tokenType = st.nextToken();
    if (tokenType == StreamTokenizer.TT_NUMBER) {
      if (compiler != VM_CompiledMethod.OPT)
        throw new Exception("optimization level can only be used with OPT compiler (int)" + st);
      optLevel = (int)st.nval;
    }
    
   
      // intern this method desscriptor
    VM_Atom cls = VM_Atom.findOrCreateAsciiAtom(className);
    VM_Atom mth = VM_Atom.findOrCreateAsciiAtom(methodName);
    VM_Atom sig = VM_Atom.findOrCreateAsciiAtom(methodSignature);

    // this is the attribute which will be returned
    VM_CompilerAdviceAttribute newAttrib;

    if (optLevel >= 0)
      newAttrib =
        new VM_CompilerAdviceAttribute(cls, mth, sig, compiler, optLevel);
    else
      newAttrib =
        new VM_CompilerAdviceAttribute(cls, mth, sig, compiler);
    
    // eat up the trailing new line
    if (tokenType != StreamTokenizer.TT_EOL) {
      tokenType = st.nextToken();
      if (tokenType != StreamTokenizer.TT_EOL)
        st.pushBack(); // don't know how to deal with it.
    }
    
    return newAttrib;
  }
}
